#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <iostream.h>
#include <sys/stat.h>
#include "mainMP.h"

/* TBitstring methods*/

TBitString::TBitString(unsigned bits, long long unsigned n) {
   this->n = n; this->bits = bits; data = new unsigned[n*bits/W+1];
};

void TBitString::Set(long long unsigned i, unsigned x) {
   SetField(data,bits,i,x);
};

unsigned TBitString::Get(long long unsigned i) {
   return GetField(data,bits,i);
};            

void TBitString::SetSize(long long unsigned n) {
   this->n = n;
};

long long unsigned TBitString::GetSize() {
   return n;
};

void TBitString::SetBits(unsigned bits) {
   this->bits = bits;
};

inline unsigned TBitString::GetBits() {
   return bits;
} 

TBitString::~TBitString() {
   delete data;
};

/* TBitStringLong methods*/

TBitStringLong::TBitStringLong(unsigned bits, long long unsigned n) {
   this->n = n; this->bits = bits; data = new unsigned[n*bits/W+1];
   if (data == NULL) { printf("not enough memory\n"); exit(1);}
};

void TBitStringLong::Set(long long unsigned i, unsigned x) {
   SetFieldLong(data,bits,i,x);
};

unsigned TBitStringLong::Get(long long unsigned i) {
   return GetFieldLong(data,bits,i);
};            

void TBitStringLong::SetSize(long long unsigned n) {
   this->n = n;
};

long long unsigned TBitStringLong::GetSize() {
   return n;
};

void TBitStringLong::SetBits(unsigned bits) {
   this->bits = bits;
};

inline unsigned TBitStringLong::GetBits() {
   return bits;
} 

TBitStringLong::~TBitStringLong() {
   delete data;
};


/* TBitSetLong methods*/

TBitSetLong::TBitSetLong(long long unsigned n) { 
   this->n = n; data = new unsigned[n/W+1]; 
};

inline void TBitSetLong::Set(long long unsigned i, bool x) {
   SetFieldLong(data,1,i,x);
};

inline bool TBitSetLong::Get(long long unsigned i) {
   return (1u << (i % W)) & data[i/W];
};

TBitSetLong_wrapper TBitSetLong::operator[] (long long unsigned i) {
   index = i; return TBitSetLong_wrapper(this);
};

TBitSetLong::~TBitSetLong() {
   delete data; 
};

/* TBitSetLong_wrapper methods*/

TBitSetLong_wrapper::TBitSetLong_wrapper(TBitSetLong *bs) {
   this->bs = bs;
};

bool TBitSetLong_wrapper::operator= (bool x) {
   bs->Set(bs->index,x); return x;
};   

TBitSetLong_wrapper::operator bool() {
   return bs->Get(bs->index);
};      

/* Mapping from DNA to alphabet {0,1,2,3} */

unsigned map(char c) 
{
   if (c=='A' || c == 'a') return 0;
   else if (c=='C' || c == 'c') return 1;
   else if (c=='G'|| c == 'g') return 2;
   else if (c=='T' || c == 't') return 3;
   else return 4;   
}

unsigned map_complement(unsigned x)
{
   if (x==0) return 3;
   else if (x==1) return 2;
   else if (x==2) return 1;
   else if (x==3) return 0;
   else return x;
}

char map_inv(unsigned x) 
{
   if (x==0) return 'A';
   else if (x==1) return 'C';
   else if (x==2) return 'G';
   else if (x==3) return 'T';
   else return '?';   
}

/* String to Index conversions*/

// sigma^x = 1u << (L*x), where L is such that sigma=2^L

inline unsigned StrToIndex(TBitStringBase *bs, long long unsigned i, unsigned k)
{
   // computes the index of string bs[i...i+k-1] among 
   // all strings of that length
   unsigned j;
   unsigned result = bs->Get(i+k-1);
   // unsigned sigma = 1u << bs->GetBits(); // size of alphabet
   for (j=0;j<k-1;j++)
      result += bs->Get(i+j)*(1u << (bs->GetBits()*(k-j-1))); 
   return result;
}

inline unsigned StrToIndex(TBitStringBase *bs, long long unsigned i, unsigned k, unsigned previndex)
{
   // computes the index of string bs[i...i+k-1] among 
   // all strings of that length, incrementally,
   // using the index of bs[i-1...i+k-2] as a starting point
   // so that the computation takes constant time 
   
   unsigned sigma = 1u << bs->GetBits();
   return sigma*(previndex-bs->Get(i-1)*(1u << (bs->GetBits()*(k-1))))+bs->Get(i+k-1);
}

inline long long unsigned StrToIndexLong(TBitStringBase *bs, long long unsigned i, unsigned k)
{
   // computes the index of string bs[i...i+k-1] among 
   // all strings of that length
   unsigned j;
   long long unsigned result = bs->Get(i+k-1);
   // unsigned sigma = 1u << bs->GetBits(); // size of alphabet
   for (j=0;j<k-1;j++)
      result += bs->Get(i+j)*(1ull << (bs->GetBits()*(k-j-1ull))); 
   return result;
}

inline long long unsigned StrToIndexLong(TBitStringBase *bs, long long unsigned i, unsigned k, long long unsigned previndex)
{
   // computes the index of string bs[i...i+k-1] among 
   // all strings of that length, incrementally,
   // using the index of bs[i-1...i+k-2] as a starting point
   // so that the computation takes constant time 
   
   long long unsigned sigma = 1u << bs->GetBits();
   return sigma*(previndex-bs->Get(i-1)*(1ull << (bs->GetBits()*(k-1))))+bs->Get(i+k-1);
}

char *IndexToStr(unsigned i, unsigned k, unsigned sbits)
{
   // computes the string corresponding 
   // to index i among all string of
   // length k from alphabet whose size is 2^sbits
   char *result = new char[k+1];
   unsigned len = 1u << (sbits*k);
   unsigned sigma = 1u << sbits;
   unsigned sigmaportion = len /sigma; 
   unsigned j=0;
   while (j < k-1) {
      result[j++] = map_inv(i / sigmaportion);
      i = i - sigmaportion*(i/sigmaportion);
      sigmaportion = sigmaportion/sigma;
   } 
   result[j++] = map_inv(i);
   result[k] = '\0';
   return result;
}

char *IndexLongToStr(long long unsigned i, unsigned k, unsigned sbits)
{
   // computes the string corresponding 
   // to index i among all string of
   // length k from alphabet whose size is 2^sbits
   char *result = new char[k+1];
   long long unsigned len = 1ull << (sbits*k);
   long long unsigned sigma = 1ull << sbits;
   long long unsigned sigmaportion = len /sigma; 
   long long unsigned j=0;
   while (j < k-1) {
      result[j++] = map_inv(i / sigmaportion);
      i = i - sigmaportion*(i/sigmaportion);
      sigmaportion = sigmaportion/sigma;
   } 
   result[j++] = map_inv(i);
   result[k] = '\0';
   return result;
}

char *Copy(char *s, unsigned l, unsigned r)
{
   char *result = new char[r-l+2];
   unsigned i;
   for (i=l;i<=r;i++)
      result[i-l] = s[i];
   result[r-l+1] = '\0';
   return result;   
}

bool FindMissingPairNaive(TBitStringBase *bs, unsigned alpha, unsigned k) 
{
   // Find missing pair (A,B), |A|=|B|=k
   // Or missing pattern for alpha=0
   bool result = false;
   unsigned sigma = 1u << bs->GetBits();
   if ((alpha>0 && 2*k>=(W+LOGW)/bs->GetBits()) || (alpha==0 && k>=(W+LOGW)/bs->GetBits())) { // word size is not enough
      printf("Maximum pointer size exceeded\n");
      printf("No missing pairs found for k<%d\n",k);
      return true; // to get out of the calling loop 
   }
   long long unsigned len;
   if (alpha>0)
      len = 1ull << (bs->GetBits()*2*k); // sigma^{2k}
   else       
      len = 1ull << (bs->GetBits()*k); // sigma^{k}
   TBitSetLong *pair = new TBitSetLong(len);
   //TBitSetLong pairp = *pair;
   long long unsigned i,j, jstart;
   long long unsigned indexi, indexj;
   for (indexi=0;indexi<len;indexi++) pair->Set(indexi,false); //pairp[indexi] = false;
   //for (indexi=0;indexi<len;indexi++) if (pairp[indexi] != false) {printf("error\n");  fflush(stdout);} 
   for (i=0;i<bs->GetSize()-k;i++) {
      if (i==0) indexi = StrToIndexLong(bs,i,k);
      else indexi = StrToIndexLong(bs,i,k,indexi);
      if (alpha==0) pair->Set(indexi,true); //pairp[indexi] = true;
      else {
         if (i<alpha) jstart = 0;
	 else jstart = i-alpha;
   	 for (j=jstart;j<=i+alpha && j<bs->GetSize()-k;j++) {
	    if (j==jstart) indexj = StrToIndexLong(bs,j,k);
	    else indexj = StrToIndexLong(bs,j,k,indexj);
	    if ((1u << (bs->GetBits()*k))*indexi+indexj>=len) {
	       printf("Index out of bounds %d>=%d\n",(1ull << (bs->GetBits()*k))*indexi+indexj,len);
	       delete pair;
	       return true;
	    }	
	    pair->Set((1u << (bs->GetBits()*k))*indexi+indexj,true);   
	    //pairp[(1u << (bs->GetBits()*k))*indexi+indexj] = true;   
	 }
      }
   }      
   if (alpha==0) {
      for (indexi=0;indexi<len;indexi++) 
         //if (pairp[indexi] == false) {
	 if (pair->Get(indexi) == false) {
	    printf("Missing pattern A=\"%s\"\n",IndexLongToStr(indexi,k,bs->GetBits()));
	    result = true;
	 }        
   } 
   else {
      for (indexi=0;indexi<len;indexi++) 
         //if (pairp[indexi] == false) {
         if (pair->Get(indexi)== false) {	 
            char *X, *A, *B;
	    X = IndexLongToStr(indexi,2*k,bs->GetBits());
	    A = Copy(X,0,k-1);
	    B = Copy(X,k,2*k-1);     
  	    printf("Missing pair (A,B)=(\"%s\",\"%s\")\n",A,B);
	    delete A; delete B; delete X;
	    result = true;
	 }        
   }
   //}
   /* Assert
   indexi = StrToIndex(bs,0,k);
   for (i=1;i<bs->n-k;i++) {
      indexi = StrToIndex(bs,i,k,indexi);
      if (StrToIndex(bs,i,k)!=indexi) {
         printf("error\n");
	 return true;
      }
   }   
   */ 
   if (!result)
      printf("No missing patterns for length k=%d\n",k);
   fflush(stdout);
   delete pair;
   return result;
}

void FindMissingPairNaive(TBitStringBase *bs, unsigned alpha) 
{
   // Find minimum k such that there is missing pair (A,B), |A|=|B|=k
   unsigned k=11;  
   printf("Starting at k=%d\n",k); 
   fflush(stdout);
   while (!FindMissingPairNaive(bs,alpha,k++));
}

int main(int argc, char *argv[])
{
   unsigned buffersize = 1u << 20;
   long long unsigned n;
   int f;
   unsigned alpha=3000;  
   if (argc==3)
      alpha = atoi(argv[1]);

   if (sizeof(unsigned)!=W/8) {
      printf("Using wrong word size W=%d, should be %d\n",W,sizeof(unsigned)*8);
      printf("Redefine W and WW accordingly, and compile again\n");
   }
   /* reading the text */
   struct stat sdata;
   stat (argv[argc-1],&sdata);
   n = (unsigned)sdata.st_size;
   printf("file size is %d MB\n",n/(1024*1024));
   n = 2*n;
   printf("file size + reverse is %d MB\n",n/(1024*1024));
fflush(stdout);
   bool usingstdin = false;
   if ((f = open(argv[argc-1], 0)) < 0){
      perror(argv[argc-1]);
      printf("Using stdin:\n");
      usingstdin = true;
   }  
   TBitStringBase *bs;    
   long long unsigned i=0,j;
   if (!usingstdin) {
      bs = new TBitStringLong(2,n);
      char *buffer = new char[buffersize];      
      buffersize = read(f,buffer,buffersize*sizeof(char));
      while (buffersize > 0) {
         j = 0;
         for (j=0;j<buffersize;j++) 
            if (map(buffer[j])<4) 
               bs->Set(i++,map(buffer[j])); 
            //else continue;
	    /*if (buffer[j]!='\n' && buffer[j]!='\0') {
               printf("Not DNA\n");
               printf("n=%d,i=%d,text[i]=%c\n",n,i,buffer[j]);
      	       return 0;
            } */ 
         buffersize = read(f,buffer,buffersize*sizeof(char));	 
      }      
      n = i;
      // read the reverse complement
      for (i=n; i<2*n;i++)
         bs->Set(i,map_complement(bs->Get(2*n-i-1)));
      n = i;
      bs->SetSize(n);
      close(f);
      delete buffer;
      printf("DNA read %d MBytes\n",n/(1024*1024));
   }
   else {
      fflush(stdout);
      bs = new TBitStringLong(2,n);
      fflush(stdout);
      char c; 
      unsigned k = 1; 
      unsigned oneprocent = (unsigned)(0.01*n/2);    
      while (cin >> c) {
         if (map(c)<4) 
            bs->Set(i++,map(c)); 
         /*else if (c!='\n' && c!='\0') {
            printf("Not DNA\n");
            printf("n=%d,i=%d,text[i]=%c\n",n,i,c);
   	    return 0;
         } */ 
	 if (i == k*oneprocent) 
  	    printf("%d \% read\n",k++);
      } 
      n = i;
      // read the reverse complement
      for (i=n; i<2*n;i++)
         bs->Set(i,map_complement(bs->Get(2*n-i-1)));
      n = i;
      bs->SetSize(n);
      printf("DNA read %d MBytes\n",n/(1024*1024));      
      fflush(stdout);
   }
   printf("Finding missing patterns for alpha=%d\n",alpha);
   fflush(stdout);
   /* Search for the missing pair */
   FindMissingPairNaive(bs,alpha);   
   delete bs;
   return 1;
}
