#define W 32
#define WW 64
#define LOGW 5

inline void SetField(unsigned *A,register unsigned len, register unsigned index,register  unsigned x) {
   unsigned i=index*len/W, 
            j=index*len-i*W;
   unsigned mask = (j+len < W ? ~0u << j+len : 0) 
                   | (W-j < W ? ~0u >> (W-j) : 0);
   A[i] = (A[i] & mask) | x << j;
   if (j+len>W) {
      mask = ((~0u) << (len+j-W));  
      A[i+1] = (A[i+1] & mask)| x >> (W-j);
   } 
}; 

inline void SetFieldLong(unsigned *A,register unsigned len, long long unsigned index,register unsigned x) {
   long long unsigned i=index*len/W, 
            j=index*len-i*W;
   unsigned mask = (j+len < W ? ~0u << j+len : 0) 
                   | (W-j < W ? ~0u >> (W-j) : 0);
   A[i] = (A[i] & mask) | x << j;
   if (j+len>W) {
      mask = ((~0u) << (len+j-W));  
      A[i+1] = (A[i+1] & mask)| x >> (W-j);
   } 
}; 

inline unsigned GetField(unsigned *A, register unsigned len, register unsigned index) {
   register unsigned i=index*len/W, 
            j=index*len-W*i,
            result;                
   if (j+len <= W)
      result = (A[i] << (W-j-len)) >> (W-len);
   else {
      result = A[i] >> j;
      result = result | (A[i+1] << (WW-j-len)) >> (W-len);
   }
   return result;
};

inline unsigned GetFieldLong(unsigned *A, register unsigned len, long long unsigned index) {
   long long unsigned i=index*len/W, 
            j=index*len-W*i;
   unsigned result;                
   if (j+len <= W)
      result = (A[i] << (W-j-len)) >> (W-len);
   else {
      result = A[i] >> j;
      result = result | (A[i+1] << (WW-j-len)) >> (W-len);
   }
   return result;
};


class TBitStringBase {
public:
    virtual void Set(long long unsigned i, unsigned x) = 0; 
    virtual unsigned Get(long long unsigned i) = 0;
    virtual void SetSize(long long unsigned n) = 0; 
    virtual long long unsigned GetSize() = 0;
    virtual void SetBits(unsigned bits) = 0; 
    virtual unsigned GetBits() = 0;
};


class TBitString : public TBitStringBase {
private:
    unsigned *data;    
    unsigned bits;
    unsigned n;
public:
    TBitString(unsigned bits, long long unsigned n); 
    inline void Set(long long unsigned i, unsigned x); 
    inline unsigned Get(long long unsigned i);
    void SetSize(long long unsigned n); 
    long long unsigned GetSize();
    void SetBits(unsigned bits); 
    inline unsigned GetBits();
    ~TBitString();
};

class TBitStringLong :public TBitStringBase {   
private:
    unsigned *data;    
    unsigned bits;
    long long unsigned n;
public:
    TBitStringLong(unsigned bits, long long unsigned n); 
    inline void Set(long long unsigned i, unsigned x); 
    inline unsigned Get(long long unsigned i);
    void SetSize(long long unsigned n); 
    long long unsigned GetSize();
    void SetBits(unsigned bits); 
    inline unsigned GetBits();
    ~TBitStringLong();
};

class TBitSetLong {
friend class TBitSetLong_wrapper;
private:
    unsigned *data;    
    long long unsigned index; 
public:
    long long unsigned n;
    TBitSetLong(long long unsigned n);
    inline void Set(long long unsigned i, bool x);
    inline bool Get(long long unsigned i);
    TBitSetLong_wrapper operator[] (long long unsigned i); 
    TBitSetLong& operator= (const bool x);
    TBitSetLong& operator= (const TBitSetLong x);
    operator bool();
    ~TBitSetLong();
};

class TBitSetLong_wrapper {
friend class TBitSetlong;
private:
   TBitSetLong *bs;
public:
   TBitSetLong_wrapper();
   TBitSetLong_wrapper(TBitSetLong *bs); 
   bool operator= (bool x); 
   operator bool();
};
